"""Azure provisioner for SkyPilot."""

from sky.provision.azure.config import bootstrap_instances
from sky.provision.azure.instance import cleanup_ports
from sky.provision.azure.instance import get_cluster_info
from sky.provision.azure.instance import open_ports
from sky.provision.azure.instance import query_instances
from sky.provision.azure.instance import run_instances
from sky.provision.azure.instance import stop_instances
from sky.provision.azure.instance import terminate_instances
from sky.provision.azure.instance import wait_instances



"""
Helper class for some OCI operations methods which needs to be shared/called 
by multiple places. 

History:
 - Hysun He (hysun.he@oracle.com) @ Apr, 2023: Initial implementation
 
"""

from datetime import datetime
import logging
import re
import time
import traceback
import typing
from typing import Optional

from sky.adaptors import common as adaptors_common
from sky.adaptors import oci as oci_adaptor
from sky.clouds.utils import oci_utils
from sky.skylet.providers.oci import utils

if typing.TYPE_CHECKING:
    import pandas as pd
else:
    pd = adaptors_common.LazyImport('pandas')

logger = logging.getLogger(__name__)


class oci_query_helper:

    # Call Cloud API to try getting the satisfied nodes.
    @classmethod
    @utils.debug_enabled(logger=logger)
    def query_instances_by_tags(cls, tag_filters, region):

        where_clause_tags = ""
        for tag_key in tag_filters:
            if where_clause_tags != "":
                where_clause_tags += " && "

            tag_value = tag_filters[tag_key]
            where_clause_tags += (f"(freeformTags.key = '{tag_key}'"
                                  f" && freeformTags.value = '{tag_value}')")

        qv_str = (f"query instance resources where {where_clause_tags}"
                  f" && (lifecycleState != 'TERMINATED'"
                  f" && lifecycleState != 'TERMINATING')")

        qv = oci_adaptor.oci.resource_search.models.StructuredSearchDetails(
            query=qv_str,
            type="Structured",
            matching_context_type=oci_adaptor.oci.resource_search.models.
            SearchDetails.MATCHING_CONTEXT_TYPE_NONE,
        )

        list_instances_response = oci_adaptor.get_search_client(
            region, oci_utils.oci_config.get_profile()).search_resources(qv)
        result_set = list_instances_response.data.items

        return result_set

    @classmethod
    def terminate_instances_by_tags(cls, tag_filters, region) -> int:
        logger.debug(f"Terminate instance by tags: {tag_filters}")
        insts = cls.query_instances_by_tags(tag_filters, region)
        fail_count = 0
        for inst in insts:
            inst_id = inst.identifier
            logger.debug(f"Got instance(to be terminated): {inst_id}")

            try:
                oci_adaptor.get_core_client(
                    region,
                    oci_utils.oci_config.get_profile()).terminate_instance(
                        inst_id)
            except Exception as e:
                fail_count += 1
                logger.error(f"Terminate instance failed: {str(e)}\n: {inst}")
                traceback.print_exc()

        if fail_count == 0:
            logger.debug(f"Instance teardown result: OK")
        else:
            logger.warn(f"Instance teardown result: {fail_count} failed!")

        return fail_count

    @classmethod
    @utils.debug_enabled(logger=logger)
    def subscribe_image(cls, compartment_id, listing_id, resource_version,
                        region):
        if (pd.isna(listing_id) or listing_id.strip() == "None" or
                listing_id.strip() == "nan"):
            return

        core_client = oci_adaptor.get_core_client(
            region, oci_utils.oci_config.get_profile())
        try:
            agreements_response = core_client.get_app_catalog_listing_agreements(
                listing_id=listing_id, resource_version=resource_version)
            agreements = agreements_response.data

            core_client.create_app_catalog_subscription(
                create_app_catalog_subscription_details=oci_adaptor.oci.core.
                models.CreateAppCatalogSubscriptionDetails(
                    compartment_id=compartment_id,
                    listing_id=listing_id,
                    listing_resource_version=agreements.
                    listing_resource_version,
                    oracle_terms_of_use_link=agreements.
                    oracle_terms_of_use_link,
                    time_retrieved=datetime.strptime(
                        re.sub(
                            "\d{3}\+\d{2}\:\d{2}",
                            "Z",
                            str(agreements.time_retrieved),
                            0,
                        ),
                        "%Y-%m-%d %H:%M:%S.%fZ",
                    ),
                    signature=agreements.signature,
                    eula_link=agreements.eula_link,
                ))
        except Exception as e:
            logger.critical(
                f"subscribe_image: {listing_id} - {resource_version} ... [Failed]"
                f"Error message: {str(e)}")
            raise RuntimeError("ERR: Image subscription error!")

    @classmethod
    @utils.debug_enabled(logger=logger)
    def find_compartment(cls, region) -> str:
        """ If compartment is not configured, we use root compartment """
        # Try to use the configured one first
        skypilot_compartment = oci_utils.oci_config.get_compartment(region)
        if skypilot_compartment is not None:
            return skypilot_compartment

        # If not specified, we try to find the one skypilot-compartment
        # Pass-in a profile parameter so that multiple profile in oci
        # config file is supported (2023/06/09).
        root = oci_adaptor.get_oci_config(
            region, oci_utils.oci_config.get_profile())['tenancy']
        list_compartments_response = oci_adaptor.get_identity_client(
            region, oci_utils.oci_config.get_profile()).list_compartments(
                compartment_id=root,
                name=oci_utils.oci_config.COMPARTMENT,
                lifecycle_state='ACTIVE',
                limit=1)
        compartments = list_compartments_response.data
        if len(compartments) > 0:
            skypilot_compartment = compartments[0].id
            return skypilot_compartment

        # Finally, we use root compartment none matches above
        skypilot_compartment = root
        return skypilot_compartment

    @classmethod
    @utils.debug_enabled(logger=logger)
    def find_create_vcn_subnet(cls, region) -> Optional[str]:
        """ If sub is not configured, we find/create VCN skypilot_vcn """
        subnet = oci_utils.oci_config.get_vcn_subnet(region)
        if subnet is not None:
            # User explicitly specified the subnet in sky config.
            return subnet

        # Try to reuse the skypilot_vcn.
        net_client = oci_adaptor.get_net_client(
            region, oci_utils.oci_config.get_profile())
        skypilot_compartment = cls.find_compartment(region)
        list_vcns_response = net_client.list_vcns(
            compartment_id=skypilot_compartment,
            display_name=oci_utils.oci_config.VCN_NAME,
            lifecycle_state="AVAILABLE")
        vcns = list_vcns_response.data
        if len(vcns) > 0:
            # Found the VCN.
            skypilot_vcn = vcns[0].id
            list_subnets_response = net_client.list_subnets(
                compartment_id=skypilot_compartment,
                limit=1,
                vcn_id=skypilot_vcn,
                display_name=oci_utils.oci_config.VCN_SUBNET_NAME,
                lifecycle_state="AVAILABLE")
            logger.debug(f'Got VCN subnet \n{list_subnets_response.data}')
            if len(list_subnets_response.data) < 1:
                logger.error(
                    f'No subnet {oci_utils.oci_config.VCN_SUBNET_NAME} '
                    f'found in the VCN {oci_utils.oci_config.VCN_NAME}')
                raise RuntimeError(
                    f'VcnSubnetNotFound Error: No subnet '
                    f'{oci_utils.oci_config.VCN_SUBNET_NAME} found in '
                    f'the VCN {oci_utils.oci_config.VCN_NAME}')
            subnet = list_subnets_response.data[0].id
            return subnet
        else:
            # Create the skypilot_vcn and related resources
            return cls.create_vcn_subnet(net_client, skypilot_compartment)

    @classmethod
    @utils.debug_enabled(logger=logger)
    def create_vcn_subnet(cls, net_client,
                          skypilot_compartment) -> Optional[str]:
        try:
            create_vcn_response = net_client.create_vcn(
                create_vcn_details=oci_adaptor.oci.core.models.CreateVcnDetails(
                    compartment_id=skypilot_compartment,
                    cidr_blocks=[oci_utils.oci_config.VCN_CIDR],
                    display_name=oci_utils.oci_config.VCN_NAME,
                    is_ipv6_enabled=False,
                    dns_label=oci_utils.oci_config.VCN_DNS_LABEL))
            vcn_data = create_vcn_response.data
            logger.debug(f'Created VCN \n{vcn_data}')
            skypilot_vcn = vcn_data.id
            route_table = vcn_data.default_route_table_id
            security_list = vcn_data.default_security_list_id
            dhcp_options_id = vcn_data.default_dhcp_options_id

            # Create internet gateway for internet access
            create_ig_response = net_client.create_internet_gateway(
                create_internet_gateway_details=oci_adaptor.oci.core.models.
                CreateInternetGatewayDetails(
                    compartment_id=skypilot_compartment,
                    is_enabled=True,
                    vcn_id=skypilot_vcn,
                    display_name=oci_utils.oci_config.VCN_INTERNET_GATEWAY_NAME
                ))
            logger.debug(
                f'Created internet gateway \n{create_ig_response.data}')
            ig = create_ig_response.data.id

            # Create a public subnet.
            create_subnet_response = net_client.create_subnet(
                create_subnet_details=oci_adaptor.oci.core.models.
                CreateSubnetDetails(
                    cidr_block=oci_utils.oci_config.VCN_SUBNET_CIDR,
                    compartment_id=skypilot_compartment,
                    vcn_id=skypilot_vcn,
                    dhcp_options_id=dhcp_options_id,
                    display_name=oci_utils.oci_config.VCN_SUBNET_NAME,
                    prohibit_internet_ingress=False,
                    prohibit_public_ip_on_vnic=False,
                    route_table_id=route_table,
                    security_list_ids=[security_list]))
            logger.debug(f'Created subnet \n{create_subnet_response.data}')
            subnet = create_subnet_response.data.id

            list_services_response = net_client.list_services(limit=100)
            services = [
                s for s in list_services_response.data
                if str(s.cidr_block).startswith('all-') and str(s.cidr_block).
                endswith('-services-in-oracle-services-network')
            ]
            if len(services) > 0:
                # Create service gateway for regional services.
                create_sg_response = net_client.create_service_gateway(
                    create_service_gateway_details=oci_adaptor.oci.core.models.
                    CreateServiceGatewayDetails(
                        compartment_id=skypilot_compartment,
                        services=[
                            oci_adaptor.oci.core.models.ServiceIdRequestDetails(
                                service_id=services[0].id)
                        ],
                        vcn_id=skypilot_vcn))
                logger.debug(f'Service Gateway: \n{create_sg_response.data}')
                sg = create_sg_response.data.id

            # Update security list: Allow all traffic in the same subnet
            update_security_list_response = net_client.update_security_list(
                security_list_id=security_list,
                update_security_list_details=oci_adaptor.oci.core.models.
                UpdateSecurityListDetails(ingress_security_rules=[
                    oci_adaptor.oci.core.models.IngressSecurityRule(
                        protocol="6",
                        source=oci_utils.oci_config.VCN_CIDR_INTERNET,
                        is_stateless=False,
                        source_type="CIDR_BLOCK",
                        tcp_options=oci_adaptor.oci.core.models.TcpOptions(
                            destination_port_range=oci_adaptor.oci.core.models.
                            PortRange(max=22, min=22),
                            source_port_range=oci_adaptor.oci.core.models.
                            PortRange(max=65535, min=1)),
                        description="Allow SSH port."),
                    oci_adaptor.oci.core.models.IngressSecurityRule(
                        protocol="all",
                        source=oci_utils.oci_config.VCN_SUBNET_CIDR,
                        is_stateless=False,
                        source_type="CIDR_BLOCK",
                        description="Allow all traffic from/to same subnet."),
                    oci_adaptor.oci.core.models.IngressSecurityRule(
                        protocol="1",
                        source=oci_utils.oci_config.VCN_CIDR_INTERNET,
                        is_stateless=False,
                        source_type="CIDR_BLOCK",
                        icmp_options=oci_adaptor.oci.core.models.IcmpOptions(
                            type=3, code=4),
                        description="ICMP traffic."),
                    oci_adaptor.oci.core.models.IngressSecurityRule(
                        protocol="1",
                        source=oci_utils.oci_config.VCN_CIDR,
                        is_stateless=False,
                        source_type="CIDR_BLOCK",
                        icmp_options=oci_adaptor.oci.core.models.IcmpOptions(
                            type=3),
                        description="ICMP traffic (VCN)."),
                ]))
            logger.debug(
                f'Updated security_list: \n{update_security_list_response.data}'
            )

            # Update route table: bind to the internet gateway
            update_route_table_response = net_client.update_route_table(
                rt_id=route_table,
                update_route_table_details=oci_adaptor.oci.core.models.
                UpdateRouteTableDetails(route_rules=[
                    oci_adaptor.oci.core.models.RouteRule(
                        network_entity_id=create_ig_response.data.id,
                        destination='0.0.0.0/0',
                        destination_type='CIDR_BLOCK',
                        description='Route table for SkyPilot VCN',
                        route_type='STATIC')
                ]))
            logger.debug(f'Route table: \n{update_route_table_response.data}')

        except oci_adaptor.service_exception() as e:
            logger.error(f'Create VCN Error: Create new VCN '
                         f'{oci_utils.oci_config.VCN_NAME} failed: {str(e)}')
            # In case of partial success while creating vcn
            cls.delete_vcn(net_client, skypilot_vcn, subnet, ig, sg)
            subnet = None

        return subnet

    @classmethod
    @utils.debug_enabled(logger=logger)
    def delete_vcn(cls, net_client, skypilot_vcn, skypilot_subnet,
                   internet_gateway, service_gateway):
        if skypilot_vcn is None:
            return  # Nothing to delete
        try:
            if internet_gateway is not None:
                # Delete internet gateway
                delete_ig_response = net_client.delete_internet_gateway(
                    ig_id=internet_gateway)
                logger.debug(f'Deleted internet gateway {internet_gateway}'
                             f'-{delete_ig_response.data}')
            if service_gateway is not None:
                # Delete service gateway
                delete_sg_response = net_client.delete_service_gateway(
                    service_gateway_id=service_gateway)
                logger.debug(f'Deleted service gateway {service_gateway}'
                             f'-{delete_sg_response.data}')
            if skypilot_subnet is not None:
                # Delete subnet
                delete_subnet_response = net_client.delete_subnet(
                    subnet_id=skypilot_subnet)
                logger.debug(f'Deleted subnet {skypilot_subnet}'
                             f'-{delete_subnet_response.data}')
            # Delete vcn
            retry_count = 0
            while retry_count < oci_utils.oci_config.MAX_RETRY_COUNT:
                try:
                    delete_vcn_response = net_client.delete_vcn(
                        vcn_id=skypilot_vcn)
                    logger.debug(
                        f'Deleted vcn {skypilot_vcn}-{delete_vcn_response.data}'
                    )
                    break
                except oci_adaptor.service_exception() as e:
                    logger.info(f'Waiting del SG/IG/Subnet finish: {str(e)}')
                    retry_count = retry_count + 1
                    if retry_count == oci_utils.oci_config.MAX_RETRY_COUNT:
                        raise e
                    else:
                        time.sleep(
                            oci_utils.oci_config.RETRY_INTERVAL_BASE_SECONDS)

        except oci_adaptor.service_exception() as e:
            logger.error(
                f'Delete VCN {oci_utils.oci_config.VCN_NAME} Error: {str(e)}')



"""Azure."""
import functools
import json
import os
import re
import subprocess
import textwrap
import typing
from typing import Any, Dict, Iterator, List, Optional, Tuple

import colorama

from sky import clouds
from sky import exceptions
from sky import sky_logging
from sky.adaptors import azure
from sky.clouds import service_catalog
from sky.utils import common_utils
from sky.utils import resources_utils
from sky.utils import ux_utils

if typing.TYPE_CHECKING:
    from sky import resources

logger = sky_logging.init_logger(__name__)

# Minimum set of files under ~/.azure that grant Azure access.
_CREDENTIAL_FILES = [
    'azureProfile.json',
    'clouds.config',
    'config',
    'msal_token_cache.json',
]

_MAX_IDENTITY_FETCH_RETRY = 10

_DEFAULT_AZURE_UBUNTU_HPC_IMAGE_GB = 30
_DEFAULT_AZURE_UBUNTU_2004_IMAGE_GB = 150


def _run_output(cmd):
    proc = subprocess.run(cmd,
                          shell=True,
                          check=True,
                          stderr=subprocess.PIPE,
                          stdout=subprocess.PIPE)
    return proc.stdout.decode('ascii')


@clouds.CLOUD_REGISTRY.register
class Azure(clouds.Cloud):
    """Azure."""

    _REPR = 'Azure'
    # Azure has a 90 char limit for resource group; however, SkyPilot adds the
    # suffix `-<region name>`. Azure also has a 64 char limit for VM names, and
    # ray adds addtional `ray-`, `-worker`, and `-<9 chars hash>` for the VM
    # names, so the limit is 64 - 4 - 7 - 10 = 43.
    # Reference: https://azure.github.io/PSRule.Rules.Azure/en/rules/Azure.ResourceGroup.Name/ # pylint: disable=line-too-long
    _MAX_CLUSTER_NAME_LEN_LIMIT = 42
    _BEST_DISK_TIER = resources_utils.DiskTier.MEDIUM
    _DEFAULT_DISK_TIER = resources_utils.DiskTier.MEDIUM
    # Azure does not support high disk tier.
    _SUPPORTED_DISK_TIERS = (set(resources_utils.DiskTier) -
                             {resources_utils.DiskTier.HIGH})

    _INDENT_PREFIX = ' ' * 4

    PROVISIONER_VERSION = clouds.ProvisionerVersion.SKYPILOT
    STATUS_VERSION = clouds.StatusVersion.SKYPILOT

    @classmethod
    def _unsupported_features_for_resources(
        cls, resources: 'resources.Resources'
    ) -> Dict[clouds.CloudImplementationFeatures, str]:
        features = {
            clouds.CloudImplementationFeatures.CLONE_DISK_FROM_CLUSTER:
                (f'Migrating disk is currently not supported on {cls._REPR}.'),
        }
        if resources.use_spot:
            features[clouds.CloudImplementationFeatures.STOP] = (
                'Stopping spot instances is currently not supported on'
                f' {cls._REPR}.')
        return features

    @classmethod
    def max_cluster_name_length(cls) -> int:
        return cls._MAX_CLUSTER_NAME_LEN_LIMIT

    def instance_type_to_hourly_cost(self,
                                     instance_type: str,
                                     use_spot: bool,
                                     region: Optional[str] = None,
                                     zone: Optional[str] = None) -> float:
        return service_catalog.get_hourly_cost(instance_type,
                                               use_spot=use_spot,
                                               region=region,
                                               zone=zone,
                                               clouds='azure')

    def accelerators_to_hourly_cost(self,
                                    accelerators: Dict[str, int],
                                    use_spot: bool,
                                    region: Optional[str] = None,
                                    zone: Optional[str] = None) -> float:
        del accelerators, use_spot, region, zone  # unused
        # Azure includes accelerators as part of the instance type.
        # Implementing this is also necessary for e.g., the instance may have 4
        # GPUs, while the task specifies to use 1 GPU.
        return 0

    def get_egress_cost(self, num_gigabytes: float):
        # In general, query this from the cloud:
        #   https://azure.microsoft.com/en-us/pricing/details/bandwidth/
        # NOTE: egress from US East.
        # NOTE: Not accurate as the pricing tier is based on cumulative monthly
        # usage.
        if num_gigabytes > 150 * 1024:
            return 0.05 * num_gigabytes
        cost = 0.0
        if num_gigabytes >= 50 * 1024:
            cost += (num_gigabytes - 50 * 1024) * 0.07
            num_gigabytes -= 50 * 1024

        if num_gigabytes >= 10 * 1024:
            cost += (num_gigabytes - 10 * 1024) * 0.083
            num_gigabytes -= 10 * 1024

        if num_gigabytes > 1:
            cost += (num_gigabytes - 1) * 0.0875

        cost += 0.0
        return cost

    @classmethod
    def get_image_size(cls, image_id: str, region: Optional[str]) -> float:
        if region is None:
            # The region used here is only for where to send the query,
            # not the image location. Azure's image is globally available.
            region = 'eastus'
        is_skypilot_image_tag = False
        if image_id.startswith('skypilot:'):
            is_skypilot_image_tag = True
            image_id = service_catalog.get_image_id_from_tag(image_id,
                                                             clouds='azure')
        image_id_splitted = image_id.split(':')
        if len(image_id_splitted) != 4:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'Invalid image id: {image_id}. Expected '
                                 'format: <publisher>:<offer>:<sku>:<version>')
        publisher, offer, sku, version = image_id_splitted
        if is_skypilot_image_tag:
            if offer == 'ubuntu-hpc':
                return _DEFAULT_AZURE_UBUNTU_HPC_IMAGE_GB
            else:
                return _DEFAULT_AZURE_UBUNTU_2004_IMAGE_GB
        compute_client = azure.get_client('compute', cls.get_project_id())
        try:
            image = compute_client.virtual_machine_images.get(
                region, publisher, offer, sku, version)
        except azure.exceptions().ResourceNotFoundError as e:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'Image not found: {image_id}') from e
        if image.os_disk_image is None:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'Retrieve image size for {image_id} failed.')
        ap = image.os_disk_image.additional_properties
        size_in_gb = ap.get('sizeInGb')
        if size_in_gb is not None:
            return float(size_in_gb)
        size_in_bytes = ap.get('sizeInBytes')
        if size_in_bytes is None:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(f'Retrieve image size for {image_id} failed. '
                                 f'Got additional_properties: {ap}')
        size_in_gb = size_in_bytes / (1024**3)
        return size_in_gb

    @classmethod
    def get_default_instance_type(
            cls,
            cpus: Optional[str] = None,
            memory: Optional[str] = None,
            disk_tier: Optional[resources_utils.DiskTier] = None
    ) -> Optional[str]:
        return service_catalog.get_default_instance_type(cpus=cpus,
                                                         memory=memory,
                                                         disk_tier=disk_tier,
                                                         clouds='azure')

    def _get_default_image_tag(self, gen_version, instance_type) -> str:
        # ubuntu-2004 v21.08.30, K80 requires image with old NVIDIA driver version
        acc = self.get_accelerators_from_instance_type(instance_type)
        if acc is not None:
            acc_name = list(acc.keys())[0]
            if acc_name == 'K80':
                return 'skypilot:k80-ubuntu-2004'

        # ubuntu-2004 v21.11.04, the previous image we used in the past for
        # V1 HyperV instance before we change default image to ubuntu-hpc.
        # In Azure, all instances with K80 (Standard_NC series), some
        # instances with M60 (Standard_NV series) and some cpu instances
        # (Basic_A, Standard_D, ...) are V1 instance. For these instances,
        # we use the previous image.
        if gen_version == 'V1':
            return 'skypilot:v1-ubuntu-2004'

        # nvidia-driver: 535.54.03, cuda: 12.2
        # see: https://github.com/Azure/azhpc-images/releases/tag/ubuntu-hpc-20230803
        # All A100 instances is of gen2, so it will always use
        # the latest ubuntu-hpc:2204 image.
        return 'skypilot:gpu-ubuntu-2204'

    @classmethod
    def regions_with_offering(cls, instance_type: str,
                              accelerators: Optional[Dict[str, int]],
                              use_spot: bool, region: Optional[str],
                              zone: Optional[str]) -> List[clouds.Region]:
        del accelerators  # unused
        assert zone is None, 'Azure does not support zones'
        regions = service_catalog.get_region_zones_for_instance_type(
            instance_type, use_spot, 'azure')

        if region is not None:
            regions = [r for r in regions if r.name == region]
        return regions

    @classmethod
    def zones_provision_loop(
        cls,
        *,
        region: str,
        num_nodes: int,
        instance_type: str,
        accelerators: Optional[Dict[str, int]] = None,
        use_spot: bool = False,
    ) -> Iterator[None]:
        del num_nodes  # unused
        regions = cls.regions_with_offering(instance_type,
                                            accelerators,
                                            use_spot,
                                            region=region,
                                            zone=None)
        for r in regions:
            assert r.zones is None, r
            yield r.zones

    # TODO: factor the following three methods, as they are the same logic
    # between Azure and AWS.

    @classmethod
    def get_accelerators_from_instance_type(
        cls,
        instance_type: str,
    ) -> Optional[Dict[str, int]]:
        return service_catalog.get_accelerators_from_instance_type(
            instance_type, clouds='azure')

    @classmethod
    def get_vcpus_mem_from_instance_type(
        cls,
        instance_type: str,
    ) -> Tuple[Optional[float], Optional[float]]:
        return service_catalog.get_vcpus_mem_from_instance_type(instance_type,
                                                                clouds='azure')

    @classmethod
    def get_zone_shell_cmd(cls) -> Optional[str]:
        return None

    def make_deploy_resources_variables(
            self,
            resources: 'resources.Resources',
            cluster_name: resources_utils.ClusterName,
            region: 'clouds.Region',
            zones: Optional[List['clouds.Zone']],
            dryrun: bool = False) -> Dict[str, Any]:
        assert zones is None, ('Azure does not support zones', zones)

        region_name = region.name

        r = resources
        # r.accelerators is cleared but .instance_type encodes the info.
        acc_dict = self.get_accelerators_from_instance_type(r.instance_type)
        acc_count = None
        if acc_dict is not None:
            custom_resources = json.dumps(acc_dict, separators=(',', ':'))
            acc_count = str(sum(acc_dict.values()))
        else:
            custom_resources = None

        if (resources.image_id is None or
                resources.extract_docker_image() is not None):
            # pylint: disable=import-outside-toplevel
            from sky.clouds.service_catalog import azure_catalog
            gen_version = azure_catalog.get_gen_version_from_instance_type(
                r.instance_type)
            image_id = self._get_default_image_tag(gen_version, r.instance_type)
        else:
            if None in resources.image_id:
                image_id = resources.image_id[None]
            else:
                assert region_name in resources.image_id, resources.image_id
                image_id = resources.image_id[region_name]
        if image_id.startswith('skypilot:'):
            image_id = service_catalog.get_image_id_from_tag(image_id,
                                                             clouds='azure')
        # Already checked in resources.py
        publisher, offer, sku, version = image_id.split(':')
        image_config = {
            'image_publisher': publisher,
            'image_offer': offer,
            'image_sku': sku,
            'image_version': version,
        }

        # Setup the A10 nvidia driver.
        need_nvidia_driver_extension = (acc_dict is not None and
                                        'A10' in acc_dict)

        # Setup commands to eliminate the banner and restart sshd.
        # This script will modify /etc/ssh/sshd_config and add a bash script
        # into .bashrc. The bash script will restart sshd if it has not been
        # restarted, identified by a file /tmp/__restarted is existing.
        # Also, add default user to docker group.
        # pylint: disable=line-too-long
        cloud_init_setup_commands = textwrap.dedent("""\
            #cloud-config
            runcmd:
              - sed -i 's/#Banner none/Banner none/' /etc/ssh/sshd_config
              - echo '\\nif [ ! -f "/tmp/__restarted" ]; then\\n  sudo systemctl restart ssh\\n  sleep 2\\n  touch /tmp/__restarted\\nfi' >> /home/skypilot:ssh_user/.bashrc
              - usermod -aG docker skypilot:ssh_user
            write_files:
              - path: /etc/apt/apt.conf.d/20auto-upgrades
                content: |
                  APT::Periodic::Update-Package-Lists "0";
                  APT::Periodic::Download-Upgradeable-Packages "0";
                  APT::Periodic::AutocleanInterval "0";
                  APT::Periodic::Unattended-Upgrade "0";
              - path: /etc/apt/apt.conf.d/10cloudinit-disable
                content: |
                  APT::Periodic::Enable "0";
            """).split('\n')

        def _failover_disk_tier() -> Optional[resources_utils.DiskTier]:
            if (r.disk_tier is not None and
                    r.disk_tier != resources_utils.DiskTier.BEST):
                return r.disk_tier
            # Failover disk tier from high to low. Default disk tier
            # (Premium_LRS, medium) only support s-series instance types,
            # so we failover to lower tiers for non-s-series.
            all_tiers = list(reversed(resources_utils.DiskTier))
            start_index = all_tiers.index(
                Azure._translate_disk_tier(r.disk_tier))
            while start_index < len(all_tiers):
                disk_tier = all_tiers[start_index]
                ok, _ = Azure.check_disk_tier(r.instance_type, disk_tier)
                if ok:
                    return disk_tier
                start_index += 1
            assert False, 'Low disk tier should always be supported on Azure.'

        return {
            'instance_type': r.instance_type,
            'custom_resources': custom_resources,
            'num_gpus': acc_count,
            'use_spot': r.use_spot,
            'region': region_name,
            # Azure does not support specific zones.
            'zones': None,
            **image_config,
            'need_nvidia_driver_extension': need_nvidia_driver_extension,
            'disk_tier': Azure._get_disk_type(_failover_disk_tier()),
            'cloud_init_setup_commands': cloud_init_setup_commands,
            'azure_subscription_id': self.get_project_id(dryrun),
            'resource_group': f'{cluster_name.name_on_cloud}-{region_name}',
        }

    def _get_feasible_launchable_resources(
        self, resources: 'resources.Resources'
    ) -> 'resources_utils.FeasibleResources':
        if resources.instance_type is not None:
            assert resources.is_launchable(), resources
            ok, _ = Azure.check_disk_tier(resources.instance_type,
                                          resources.disk_tier)
            if not ok:
                # TODO: Add hints to all return values in this method to help
                #  users understand why the resources are not launchable.
                return resources_utils.FeasibleResources([], [], None)
            # Treat Resources(Azure, Standard_NC4as_T4_v3, T4) as
            # Resources(Azure, Standard_NC4as_T4_v3).
            resources = resources.copy(accelerators=None)
            return resources_utils.FeasibleResources([resources], [], None)

        def _make(instance_list):
            resource_list = []
            for instance_type in instance_list:
                ok, _ = Azure.check_disk_tier(instance_type,
                                              resources.disk_tier)
                if not ok:
                    continue
                r = resources.copy(
                    cloud=Azure(),
                    instance_type=instance_type,
                    # Setting this to None as Azure doesn't separately bill /
                    # attach the accelerators.  Billed as part of the VM type.
                    accelerators=None,
                    cpus=None,
                    memory=None,
                )
                resource_list.append(r)
            return resource_list

        # Currently, handle a filter on accelerators only.
        accelerators = resources.accelerators
        if accelerators is None:
            # Return a default instance type with the given number of vCPUs.
            default_instance_type = Azure.get_default_instance_type(
                cpus=resources.cpus,
                memory=resources.memory,
                disk_tier=resources.disk_tier)
            if default_instance_type is None:
                return resources_utils.FeasibleResources([], [], None)
            else:
                return resources_utils.FeasibleResources(
                    _make([default_instance_type]), [], None)

        assert len(accelerators) == 1, resources
        acc, acc_count = list(accelerators.items())[0]
        (instance_list, fuzzy_candidate_list
        ) = service_catalog.get_instance_type_for_accelerator(
            acc,
            acc_count,
            cpus=resources.cpus,
            memory=resources.memory,
            use_spot=resources.use_spot,
            region=resources.region,
            zone=resources.zone,
            clouds='azure')
        if instance_list is None:
            return resources_utils.FeasibleResources([], fuzzy_candidate_list,
                                                     None)
        return resources_utils.FeasibleResources(_make(instance_list),
                                                 fuzzy_candidate_list, None)

    @classmethod
    def check_credentials(cls) -> Tuple[bool, Optional[str]]:
        """Checks if the user has access credentials to this cloud."""
        help_str = (
            ' Run the following commands:'
            f'\n{cls._INDENT_PREFIX}  $ az login'
            f'\n{cls._INDENT_PREFIX}  $ az account set -s <subscription_id>'
            f'\n{cls._INDENT_PREFIX}For more info: '
            'https://docs.microsoft.com/en-us/cli/azure/get-started-with-azure-cli'  # pylint: disable=line-too-long
        )
        # This file is required because it will be synced to remote VMs for
        # `az` to access private storage buckets.
        # `az account show` does not guarantee this file exists.
        azure_token_cache_file = '~/.azure/msal_token_cache.json'
        if not os.path.isfile(os.path.expanduser(azure_token_cache_file)):
            return (False,
                    f'{azure_token_cache_file} does not exist.' + help_str)

        try:
            _run_output('az --version')
        except subprocess.CalledProcessError as e:
            return False, (
                # TODO(zhwu): Change the installation hint to from PyPI.
                'Azure CLI `az --version` errored. Run the following commands:'
                f'\n{cls._INDENT_PREFIX}  $ pip install skypilot[azure]'
                f'\n{cls._INDENT_PREFIX}Credentials may also need to be set.'
                f'{help_str}\n'
                f'{cls._INDENT_PREFIX}Details: '
                f'{common_utils.format_exception(e)}')
        # If Azure is properly logged in, this will return the account email
        # address + subscription ID.
        try:
            cls.get_current_user_identity()
        except exceptions.CloudUserIdentityError as e:
            return False, (f'Getting user\'s Azure identity failed.{help_str}\n'
                           f'{cls._INDENT_PREFIX}Details: '
                           f'{common_utils.format_exception(e)}')

        # Check if the azure blob storage dependencies are installed.
        try:
            # pylint: disable=redefined-outer-name, import-outside-toplevel, unused-import
            from azure.storage import blob
            import msgraph
        except ImportError as e:
            return False, (
                f'Azure blob storage depdencies are not installed. '
                'Run the following commands:'
                f'\n{cls._INDENT_PREFIX}  $ pip install skypilot[azure]'
                f'\n{cls._INDENT_PREFIX}Details: '
                f'{common_utils.format_exception(e)}')
        return True, None

    def get_credential_file_mounts(self) -> Dict[str, str]:
        """Returns a dict of credential file paths to mount paths."""
        return {
            f'~/.azure/{filename}': f'~/.azure/{filename}'
            for filename in _CREDENTIAL_FILES
        }

    def instance_type_exists(self, instance_type):
        return service_catalog.instance_type_exists(instance_type,
                                                    clouds='azure')

    @classmethod
    @functools.lru_cache(maxsize=1)  # Cache since getting identity is slow.
    def get_current_user_identity(cls) -> Optional[List[str]]:
        """Returns the cloud user identity."""
        # This returns the user's email address + [subscription_id].
        retry_cnt = 0
        while True:
            retry_cnt += 1
            try:
                import knack  # pylint: disable=import-outside-toplevel
            except ModuleNotFoundError as e:
                with ux_utils.print_exception_no_traceback():
                    raise exceptions.CloudUserIdentityError(
                        'Failed to import \'knack\'. To install the dependencies for Azure, '
                        'Please install SkyPilot with: '
                        f'{colorama.Style.BRIGHT}pip install skypilot[azure]'
                        f'{colorama.Style.RESET_ALL}') from e
            try:
                account_email = azure.get_current_account_user()
                break
            except (FileNotFoundError, knack.util.CLIError) as e:
                error = exceptions.CloudUserIdentityError(
                    'Failed to get activated Azure account.\n'
                    '  Reason: '
                    f'{common_utils.format_exception(e, use_bracket=True)}')
                if retry_cnt <= _MAX_IDENTITY_FETCH_RETRY:
                    logger.debug(f'{error}.\nRetrying...')
                    continue
                with ux_utils.print_exception_no_traceback():
                    raise error from None
            except Exception as e:  # pylint: disable=broad-except
                with ux_utils.print_exception_no_traceback():
                    raise exceptions.CloudUserIdentityError(
                        'Failed to get Azure user identity with unknown '
                        f'exception.\n'
                        '  Reason: '
                        f'{common_utils.format_exception(e, use_bracket=True)}'
                    ) from e
        try:
            project_id = cls.get_project_id()
        except (ModuleNotFoundError, RuntimeError) as e:
            with ux_utils.print_exception_no_traceback():
                raise exceptions.CloudUserIdentityError(
                    'Failed to get Azure project ID.') from e
        return [f'{account_email} [subscription_id={project_id}]']

    @classmethod
    def get_current_user_identity_str(cls) -> Optional[str]:
        user_identity = cls.get_current_user_identity()
        if user_identity is None:
            return None
        return user_identity[0]

    @classmethod
    def get_project_id(cls, dryrun: bool = False) -> str:
        if dryrun:
            return 'dryrun-project-id'
        try:
            azure_subscription_id = azure.get_subscription_id()
            if not azure_subscription_id:
                raise ValueError  # The error message will be replaced.
        except ModuleNotFoundError as e:
            with ux_utils.print_exception_no_traceback():
                raise ModuleNotFoundError('Unable to import azure python '
                                          'module. Is azure-cli python package '
                                          'installed? Try pip install '
                                          '.[azure] in the sky repo.') from e
        except Exception as e:  # pylint: disable=broad-except
            with ux_utils.print_exception_no_traceback():
                raise RuntimeError(
                    'Failed to get subscription id from azure cli. '
                    'Make sure you have logged in and run this Azure '
                    'cli command: "az account set -s <subscription_id>".'
                ) from e
        return azure_subscription_id

    @classmethod
    def _is_s_series(cls, instance_type: Optional[str]) -> bool:
        # For azure naming convention, see https://learn.microsoft.com/en-us/azure/virtual-machines/vm-naming-conventions  # pylint: disable=line-too-long
        if instance_type is None:
            return True
        x = re.match(
            r'(Standard|Basic)_([A-Z]+)([0-9]+)(-[0-9]+)?'
            r'([a-z]*)(_[A-Z]+[0-9]+)?(_v[0-9])?(_Promo)?', instance_type)
        assert x is not None, f'Unknown instance type: {instance_type}'
        return 's' in x.group(5)

    @classmethod
    def check_disk_tier(
            cls, instance_type: Optional[str],
            disk_tier: Optional[resources_utils.DiskTier]) -> Tuple[bool, str]:
        if disk_tier is None or disk_tier == resources_utils.DiskTier.BEST:
            return True, ''
        if disk_tier == resources_utils.DiskTier.HIGH:
            return False, ('Azure disk_tier=high is not supported now. '
                           'Please use disk_tier={low, medium} instead.')
        # Only S-series supported premium ssd
        # see https://stackoverflow.com/questions/48590520/azure-requested-operation-cannot-be-performed-because-storage-account-type-pre  # pylint: disable=line-too-long
        if cls._get_disk_type(
                disk_tier
        ) == 'Premium_LRS' and not Azure._is_s_series(instance_type):
            return False, (
                'Azure premium SSDs are only supported for S-series '
                'instances. To use disk_tier=medium, please make sure '
                'instance_type is specified to an S-series instance.')
        return True, ''

    @classmethod
    def check_disk_tier_enabled(cls, instance_type: Optional[str],
                                disk_tier: resources_utils.DiskTier) -> None:
        ok, msg = cls.check_disk_tier(instance_type, disk_tier)
        if not ok:
            with ux_utils.print_exception_no_traceback():
                raise exceptions.NotSupportedError(msg)

    @classmethod
    def _get_disk_type(cls,
                       disk_tier: Optional[resources_utils.DiskTier]) -> str:
        tier = cls._translate_disk_tier(disk_tier)
        # TODO(tian): Maybe use PremiumV2_LRS/UltraSSD_LRS? Notice these two
        # cannot be used as OS disks so we might need data disk support
        tier2name = {
            resources_utils.DiskTier.HIGH: 'Disabled',
            resources_utils.DiskTier.MEDIUM: 'Premium_LRS',
            resources_utils.DiskTier.LOW: 'Standard_LRS',
        }
        return tier2name[tier]
